/*
   cobol.c - COBOL language functions for deckmate program.

   Author: Paul Hardy, June 2020

   Copyright (C) 2020 Paul Hardy

   LICENSE:

      This program is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "deckmate.h"


/*
   Draw invariant form markings for COBOL listings.
*/
void
new_ps_page_cobol (FILE *outfp) {

   /*
       Set tabs for COBOL Page/Line border (column 4)
       and Margin B (column 12) to print dashed gray
       lines down the page.
   */
   int tab_stops[3]={4,12,0};

   void new_ps_page_tabs (FILE *outfp, int *tab_stops, int);


   fprintf (outfp, "   %%\n");
   fprintf (outfp, "   %% Template for COBOL listings.\n");
   fprintf (outfp, "   %%\n");

   /* Print Sequence field: 6 columns */
   fprintf (outfp, "   /Helvetica findfont 5 scalefont setfont\n");

   fprintf (outfp, "   %d %d moveto 0 %d rlineto stroke\n",
           LEFT_MARGIN_80 + 36, TOP_MARGIN_80 + 20, -22);

   /* Center text over columns 1-6 */
   fprintf (outfp, "   (SEQUENCE) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 18, TOP_MARGIN_80 + 34);
   fprintf (outfp, "   show\n");
   fprintf (outfp, "   (NUMBER) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 18, TOP_MARGIN_80 + 28);
   fprintf (outfp, "   show\n");
   fprintf (outfp, "   (\\(PAGE\\)) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 9, TOP_MARGIN_80 + 22);
   fprintf (outfp, "   show\n");
   fprintf (outfp, "   (\\(LINE\\)) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 27, TOP_MARGIN_80 + 22);
   fprintf (outfp, "   show\n");

   /* Print Continuation field: 1 column */
   fprintf (outfp, "   %d %d moveto 0 %d rlineto stroke\n",
           LEFT_MARGIN_80 + 42, TOP_MARGIN_80 + 20, -22);

   fprintf (outfp, "   %d %d moveto 90 rotate\n",
           LEFT_MARGIN_80 + 41, TOP_MARGIN_80 + 22);
   fprintf (outfp, "   (CONT.) show\n");
   fprintf (outfp, "   -90 rotate\n");
 
   /* Print COBOL Page/Line tab and Margin B tab before column 12 */
   new_ps_page_tabs (outfp, tab_stops, 0);

   /* Print Program Identification field: 8 columns */
   /* Center text on line */
   fprintf (outfp, "   (PROGRAM) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 456, TOP_MARGIN_80 + 28);
   fprintf (outfp, "   show\n");
   fprintf (outfp, "   (IDENTIFICATION) dup stringwidth pop\n");
   fprintf (outfp, "   -2 div %d add %d moveto\n",
            LEFT_MARGIN_80 + 456, TOP_MARGIN_80 + 22);
   fprintf (outfp, "   show\n");

   fprintf (outfp, "   /Helvetica findfont 10 scalefont setfont\n");

   /* Center text on line, adding 3 pt extra for space character (32) */
   fprintf (outfp, "   3 0 32 3 0");
   fprintf (outfp, " (COBOL STATEMENT) dup stringwidth pop\n");
   /*
      Adjust starting point to be centered on top of form,
      allowing for 3 pt kerning between 14 letter pairs.
   */
   fprintf (outfp, "   -2 div %f add %d moveto\n",
           LEFT_MARGIN_80 + BOX_WIDTH_80 / 2.0 - 3 * 16 / 2 + 0.5,
           TOP_MARGIN_80 + 22);
   fprintf (outfp, "   awidthshow\n");

   /* Solid line down the form for comments at column 73 */
   fprintf (outfp, "   %d %d moveto 0 %d rlineto stroke\n",
            LEFT_MARGIN_80 + 432, TOP_MARGIN_80 + 20, -BOX_HEIGHT_80 - 20);

   return;
}


/*
   Print context-dependent vertical rules on a line for COBOL.
*/
void
ps_line_out_cobol (FILE *outfp,
                   int line_num, char *inbuffer, language form_lang) {

   int i;  /* loop variable */

   int *verticals;  /* Points to one of the arrays of verticals below */
   int default_verticals[81] = {6, 7, 0};
   int comment_verticals[81] = {6, 7, 0};
   int jcl_verticals    [81] = {0};


   verticals = NULL;  /* No matching pattern found yet. */

   /*
      Print inter-field vertical lines if not a comment line.
   */
   if ((inbuffer [0] == '/') || (inbuffer [1] == '/')) { /* If IBM JCL */
         verticals = jcl_verticals;
   }
   else if (inbuffer [6] == '*') {
         verticals = comment_verticals;
   }
   else {
         verticals = default_verticals;
   }
   if (verticals != NULL) {
      for (i = 0; verticals[i] != 0; i++) {
         fprintf (outfp, "%d %d tick\n",
                 LEFT_MARGIN_80 + 6 * verticals[i],
                 TOP_MARGIN_80 + 8 - 10 * line_num);
      }
   }

   return;
}

