/*
   input.c - functions to process deckmate input text strings for output.

   Author: Paul Hardy, June 2020

   Copyright (C) 2020 Paul Hardy

   LICENSE:

      This program is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
      the Free Software Foundation, either version 2 of the License, or
      (at your option) any later version.

      This program is distributed in the hope that it will be useful,
      but WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
      along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "deckmate.h"


/*
   Determine next line number from carriage control code
   that was in the first column of the current input line.
*/
int
get_next_line (char cc_char, int line_num, int init,
               int fortran_cc, int *cc_tape) {

   int next_line;


   next_line = line_num;
  
   /* Handle FORTRAN carriage control, if option is selected */
   if (fortran_cc == 1) {
      switch (cc_char) {
         case ' ': if (init == 1) {  /* If at top of 1st page */
                      next_line = cc_tape [0];
                   }
                   else {
                      next_line++;
                   }
                   break;   /* Single space      */
         case '0': next_line +=  2; break;   /* Double space      */
         case '-': next_line +=  3; break;   /* Triple space      */
         case '1': /* Advance to next page if not at start of printout */
         case '\f':                         /* <FF> --> new page */
                   if (init == 1) {  /* At the start of the 1st page */
                      next_line  = cc_tape [0];
                   }
                   else {
                      next_line = 66 + cc_tape[0];
                   }
                   break;
         case '\v':next_line +=  6 - (next_line-1) % 6; /* <VT> on 6 lines */
                   break;
         case '+':                 break;   /* No line advance   */
         default:  next_line++;    break;   /* Single space      */
      }
   }
  

   return next_line;
}


/*
   Build line formatted for PostScript output.
*/
void
build_line (char *inbuffer, int chars_per_line, int fortran_cc,
            char *safe_buffer, int *symbol, unsigned *fonts_used) {

   int  i, j, k;     /* loop variables      */
   int  output_col;   /* current column in output line */

   /* Skip over the carriage control format character if enabled */
   i = 0;
   if (fortran_cc == 1 || inbuffer[0] == 0xB || inbuffer[0] == 0xC) i = 1;
   j = 0;
   /* Line is started.  Process one character at a time. */
   output_col = 1;
   while (output_col  <= chars_per_line              &&
          inbuffer[i] != '\r' && inbuffer[i] != '\n' &&
          inbuffer[i] != '\0') {

      /* Output spaces for one or more tabs and/or spaces in sequence */
      while (inbuffer[i] == '\t' && output_col <= chars_per_line) {

         k = 8 - ((output_col - 1) % 8);  /* spaces to add */
         output_col += k;
         while (k-- > 0) safe_buffer[j++] = ' ';

         i++;
      }

      /* Escape some characters for well-formed PostScript strings. */
      if (inbuffer[i] == '(' || inbuffer[i] == ')' ||
          inbuffer[i] == '/' || inbuffer[i] == '\\') {
         safe_buffer[j++] = '\\';
         safe_buffer[j++] = inbuffer[i];
         output_col++;
         i++;
      }
      else if (inbuffer[i] < 0x20 || inbuffer[i] >= 0x7F) {
         /*
            Check for "cents" symbol (U+00A2), as a single
            binary byte value or as UTF-8.
         */
         if ((inbuffer[i] == '\242') ||
             (inbuffer[i] == '\302' && inbuffer[i+1] == '\242')) {
            strcpy (&safe_buffer[j], "\\242");  /* cents symbol */
            j += 4;
            output_col++;
            if (inbuffer[i] == '\302') i++;  /* if two-byte UTF-8 */
            i++;
         }
         /*
            Check for "logical not" symbol (U+00AC), as a single
            binary byte value or as UTF-8.
         */
         else if ((inbuffer[i] == '\254') ||
                  (inbuffer[i] == '\302' && inbuffer[i+1] == '\254')) {
            /*
               Logical not symbol; flag as a PostScript font dagger
               glyph for ps_line_out function to replace with the
               logical not glyph from PostScript Symbol font.
            */
            /* '\330' will be printed as the Symbol logical not glyph */
            strcpy (&safe_buffer[j], "\\330");
            j += 4;
            output_col++;
            if (inbuffer[i] == '\302') i++;  /* if two-byte UTF-8 */
            i++;
            *symbol = 1;  /* mark output string as containing this glyph */
            *fonts_used |= FONT_PS_SYMBOL;
         }
         else {
            /*
               Unknown or unprintable character; print a double dagger.
            */
            strcpy (&safe_buffer[j], "\\263");
            j += 4;
            output_col++;
            /*
               If this looks like a valid UTF-8 code point in the range
               U+0080..U+00FF (\302\200..\302\277), treat this input
               byte and the one that follows as one character.
            */
            if (inbuffer[i] == '\302') {
               if (inbuffer[i+1] >= '\200' && inbuffer[i+1] <= '\277')
                  i += 2;
               else
                  i++;
            }
         }  /* Unknown or unprintable character; print a double dagger. */
      }  /* Byte is in range U+0000..U+001F or U+0080..U+00FF */
      else {
         safe_buffer[j++] = inbuffer[i];
         output_col++;
         i++;
      }
   }
   safe_buffer[j] = '\0';


   return;
}

